from .base_phantom import BasePhantom
from ..elements import Sphere


class SimpleCell(BasePhantom):
    def __init__(self,
                 cytoplasm_index=1.365,
                 nucleus_index=1.360,
                 nucleoli_index=1.387,
                 medium_index=1.335,
                 cytoplasm_fl=0,
                 nucleus_fl=3,
                 nucleus_shell_fl=10,
                 nucleoli_fl=0):
        """Simple cell phantom with cytoplasm, nucleus, and 2 nucleoli

        Configuration:
        - cytoplasm_index sphere (no fl) containing
          - nucleus sphere (shell fluorescence) containing
            - 2 small nucleoli (full fluorescence)
        """
        nleoi1 = Sphere(radius=1.5e-6,
                        center=(-.25e-6, 2e-6, 2e-6),
                        fl_brightness=nucleoli_fl-nucleus_fl,
                        object_index=medium_index +
                        (nucleoli_index-nucleus_index),
                        medium_index=medium_index,
                        )

        nleoi2 = Sphere(radius=1.5e-6,
                        center=(+.25e-6, -1e-6, 2e-6),
                        fl_brightness=nucleoli_fl-nucleus_fl,
                        object_index=medium_index +
                        (nucleoli_index-nucleus_index),
                        medium_index=medium_index,
                        )

        nuclus = Sphere(radius=4e-6,
                        center=(0e-6, 1e-6, 1e-6),
                        fl_brightness=nucleus_shell_fl,
                        object_index=medium_index +
                        (nucleus_index-cytoplasm_index),
                        medium_index=medium_index,
                        )

        nuclus_shell = Sphere(radius=3.8e-6,
                              center=(0e-6, 1e-6, 1e-6),
                              fl_brightness=nucleus_fl-nucleus_shell_fl,
                              object_index=medium_index,
                              medium_index=medium_index,
                              )

        cyto = Sphere(radius=5.5e-6,
                      center=(0, 0, 0),
                      fl_brightness=cytoplasm_fl,
                      object_index=cytoplasm_index,
                      medium_index=medium_index,
                      )
        self.elements = [nleoi1, nleoi2, nuclus, nuclus_shell, cyto]
        self.medium_index = medium_index



import os
import os.path as op
import sys

import matplotlib.pylab as plt

thisdir = op.dirname(op.abspath(__file__))
sys.path.insert(0, op.dirname(thisdir))

DPI = 80


if __name__ == "__main__":
    # Do not display example plots
    plt.show = lambda: None
    files = os.listdir(thisdir)
    files = [f for f in files if f.endswith(".py")]
    files = [f for f in files if not f == op.basename(__file__)]
    files = sorted([op.join(thisdir, f) for f in files])

    for f in files:
        fname = f[:-3] + ".png"
        if not op.exists(fname):
            exec_str = open(f).read()
            if exec_str.count("plt.show()"):
                exec(exec_str)
                plt.savefig(fname, dpi=DPI)
                print("Image created: '{}'".format(fname))
            else:
                print("No image: '{}'".format(fname))
        else:
            print("Image skipped (already exists): '{}'".format(fname))
        plt.close()



import pathlib

import h5py
import numpy as np
import qpimage

from .fluorescence import Fluorescence
from .propagators import dictionary as pp_dict


class Sinogram(object):
    def __init__(self, phantom, wavelength, pixel_size, grid_size):
        self.phantom = phantom
        self.wavelength = wavelength
        self.pixel_size = pixel_size
        self.grid_size = grid_size

    def compute(self, angles, path=None, propagator="rytov"):
        if isinstance(angles, int):
            angles = np.linspace(0, 2*np.pi, angles, endpoint=False)

        if path:
            write = True
            path = pathlib.Path(path)
            if path.exists():
                path.unlink()
        else:
            write = False
            sino_fields = np.zeros((angles.size,
                                    self.grid_size[0],
                                    self.grid_size[1]),
                                   dtype=complex)
            sino_fluor = np.zeros((angles.size,
                                   self.grid_size[0],
                                   self.grid_size[1]),
                                  dtype=float)

        for ii, ang in enumerate(angles):
            ph = self.phantom.transform(rot_main=ang)

            pp = pp_dict[propagator](phantom=ph,
                                     grid_size=self.grid_size,
                                     pixel_size=self.pixel_size,
                                     wavelength=self.wavelength)
            qpi = pp.propagate()
            fluor = Fluorescence(phantom=ph,
                                 grid_size=self.grid_size,
                                 pixel_size=self.pixel_size).project()

            if write:
                with h5py.File(path, "a") as h5:
                    qps = qpimage.QPSeries(h5file=h5.require_group("qpseries"))
                    qps.add_qpimage(qpi)

                    h5fl = h5.require_group("fluorescence")
                    h5fl.create_dataset("{}".format(ii), data=fluor)
            else:
                sino_fields[ii] = qpi.field
                sino_fluor[ii] = fluor

        if write:
            return path
        else:
            return sino_fields, sino_fluor



import numpy as np
import qpsphere

from .base_propagator import BasePropagator


class Rytov(BasePropagator):
    """Rytov approximation"""

    def propagate_sphere(self, sphere, grid_sampling=150):
        center = self.center + sphere.points[0]/self.pixel_size
        # speed up computation for smaller spheres on large grid
        n = np.max(self.grid_size) * self.pixel_size / sphere.radius
        radius_sampling = grid_sampling / n
        qpi = qpsphere.models.rytov(radius=sphere.radius,
                                    sphere_index=sphere.object_index,
                                    medium_index=sphere.medium_index,
                                    wavelength=self.wavelength,
                                    pixel_size=self.pixel_size,
                                    grid_size=self.grid_size,
                                    center=center[:2],
                                    focus=-center[2]*self.pixel_size,
                                    radius_sampling=radius_sampling,
                                    )
        return qpi



import numpy as np

from .elements import Sphere


class Fluorescence(object):

    def __init__(self, phantom, grid_size, pixel_size):
        """Fluorescence projector

        Notes
        -----
        - The origin of the coordinate system is the center of the grid.
          For even grid sizes, the origin is between two pixels. For odd
          grid sizes the origin coincides with the center pixel.
        """
        self.phantom = phantom
        self.pixel_size = pixel_size
        self.grid_size = grid_size
        gx, gy = grid_size
        #: center of the volume used
        self.center = np.array([gx, gy, 0]) / 2 - .5

    def project(self):
        fluor = np.zeros(self.grid_size, dtype=float)
        for element in self.phantom:
            if isinstance(element, Sphere):
                fluor += self.project_sphere(element)

        return fluor

    def project_sphere(self, sphere):
        center = self.center + sphere.center/self.pixel_size
        # grid
        x = np.arange(self.grid_size[0]).reshape(-1, 1)
        y = np.arange(self.grid_size[1]).reshape(1, -1)
        cx, cy, _ = center
        # sphere location
        rpx = sphere.radius / self.pixel_size
        r = rpx**2 - (x - cx)**2 - (y - cy)**2
        # distance
        z = np.zeros_like(r)
        rvalid = r > 0
        z[rvalid] = 2 * np.sqrt(r[rvalid])

        fluor = z * sphere.fl_brightness
        return fluor



import copy

import numpy as np


class BaseElement(object):
    def __init__(self, object_index, medium_index, fl_brightness,
                 points):
        """Initialize a basic element

        Parameters
        ----------
        object_index: float
            Refractive index of the element
        medium_index: float
            Refractive index of surrounding medium
        fl_brightness: float
            Fluorescence brightness
        points: 2d ndarray
            Coordinates of the element [m]

        Notes
        -----
        When subclassing this class, override this method with
        additional parameters (e.g. position, size) and call it
        with super(ClassName, self).__init__(...).
        """
        #: refractive index of the object
        self.object_index = object_index
        #: refractive index of the medium
        self.medium_index = medium_index
        #: brightness of the fluorescence signal
        self.fl_brightness = fl_brightness
        #: 2D array of points describing the geometrical object. This
        #: variable is used for affine transforms (e.g. when rotating
        #: the object).
        self.points = np.array(points)

    def draw(self, grid_size, pixel_size):
        ri = np.ones(grid_size, dtype=float) * self.medium_index
        fl = np.zeros(grid_size, dtype=float)
        for pp in self.points:
            # ODTbrain convention
            cy, cz, cx = np.array(pp/pixel_size, dtype=int)
            ri[cx, cy, cz] = self.object_index
            fl[cx, cy, cz] = self.fl_brightness
        return ri, fl

    def transform(self, x=0, y=0, z=0, rot_main=0, rot_in_plane=0,
                  rot_perp_plane=0):
        """Rotate and translate self.points

        Notes
        -----
        - By convention, sinogram generation in cellsino is performed by
          modifying the pitch (rotation about y-axis).
        - Rotation is performed prior to translation. First, the points
          are rotated about the y-axis (``rot_main``, the main sinogram
          acquisition angle). Second, the points are rotated about the
          x-axis (``rot_perp_plane``, perpendicular to the imaging plane).
          Third, the points are rotated about the z-axis (``rot_in_plane``,
          within the imaging plane).
        """
        # In ODTbrain convention, x -> -y and y -> x.
        alpha = -rot_main
        beta = rot_perp_plane
        gamma = rot_in_plane
        Rx = np.array([
                      [1,             0,              0],
                      [0, np.cos(alpha), -np.sin(alpha)],
                      [0, np.sin(alpha),  np.cos(alpha)],
                      ])

        Ry = np.array([
                      [np.cos(beta),  0, np.sin(beta)],
                      [0,             1,            0],
                      [-np.sin(beta), 0, np.cos(beta)],
                      ])

        Rz = np.array([
                      [np.cos(gamma), -np.sin(gamma), 0],
                      [np.sin(gamma),  np.cos(gamma), 0],
                      [0,          0,                 1],
                      ])

        R = np.dot(np.dot(Ry, Rz), Rx)
        rotated = np.dot(R, self.points.T).T
        rotated_pad = np.pad(rotated, ((0, 0), (0, 1)), mode="constant",
                             constant_values=1)

        T = np.array([[1, 0, 0, x],
                      [0, 1, 0, y],
                      [0, 0, 1, z],
                      [0, 0, 0, 1],
                      ])
        translated = np.dot(T, rotated_pad.T)[:-1].T

        # return a copy of the current instance with points transformed
        telement = copy.copy(self)
        telement.points = translated
        return telement



import qpsphere

from .base_propagator import BasePropagator


class Projection(BasePropagator):
    """Projection approximation"""

    def propagate_sphere(self, sphere):
        center = self.center + sphere.center/self.pixel_size
        qpi = qpsphere.models.projection(radius=sphere.radius,
                                         sphere_index=sphere.object_index,
                                         medium_index=sphere.medium_index,
                                         wavelength=self.wavelength,
                                         pixel_size=self.pixel_size,
                                         grid_size=self.grid_size,
                                         center=center[:2],
                                         )
        return qpi



import numpy as np

from .base_element import BaseElement


class Sphere(BaseElement):
    def __init__(self, object_index, medium_index, fl_brightness,
                 center, radius):
        """Sphere element

        Parameters
        ----------
        object_index: float
            Refractive index of the element
        medium_index: float
            Refractive index of surrounding medium
        fl_brightness: float
            Fluorescence brightness
        center: list-like
            Center coordinates (x, y, z) of the sphere [m]
        radius:
            Radius of the sphere [m]
        """
        #: radius of the sphere
        self.radius = radius
        points = np.atleast_2d(center)
        super(Sphere, self).__init__(object_index=object_index,
                                     medium_index=medium_index,
                                     fl_brightness=fl_brightness,
                                     points=points)

    @property
    def center(self):
        return self.points[0]

    def draw(self, grid_size, pixel_size):
        ri = np.ones(grid_size, dtype=float) * self.medium_index
        fl = np.zeros(grid_size, dtype=float)
        center = np.array(grid_size) / 2 - .5
        x = (np.arange(grid_size[0]) - center[0]) * pixel_size
        y = (np.arange(grid_size[1]) - center[1]) * pixel_size
        z = (np.arange(grid_size[2]) - center[2]) * pixel_size

        xx, yy, zz = np.meshgrid(x, y, z, indexing="ij", sparse=True)
        cy, cz, cx = self.center

        volume = (cx-xx)**2 + (cy-yy)**2 + (cz-zz)**2
        inside = volume <= self.radius**2
        ri[inside] = self.object_index
        fl[inside] = self.fl_brightness
        return ri, fl



import numpy as np


class BasePhantom(object):
    def __init__(self, medium_index):
        self.elements = []
        self.medium_index = medium_index

    def __iter__(self):
        for el in self.elements:
            yield el

    def append(self, element):
        self.elements.append(element)

    def draw(self, grid_size, pixel_size):
        ri = np.ones(grid_size, dtype=float) * self.medium_index
        fl = np.zeros(grid_size, dtype=float)

        for el in self:
            riel, flel = el.draw(grid_size, pixel_size)
            ri += riel - el.medium_index
            fl += flel
        return ri, fl

    def transform(self, x=0, y=0, z=0, rot_main=0, rot_in_plane=0,
                  rot_perp_plane=0):
        ph = BasePhantom(medium_index=self.medium_index)
        for el in self:
            ph.append(el.transform(x=x,
                                   y=y,
                                   z=z,
                                   rot_main=rot_main,
                                   rot_in_plane=rot_in_plane,
                                   rot_perp_plane=rot_perp_plane))
        return ph



from os.path import dirname, realpath, exists
from setuptools import setup, find_packages
import sys


author = u"Paul Müller"
authors = [author]
description = 'in-silico sinograms (phase and fluorescence) of cells'
name = 'cellsino'
year = "2019"

sys.path.insert(0, realpath(dirname(__file__))+"/"+name)
from _version import version

setup(
    name=name,
    author=author,
    author_email='dev@craban.de',
    url='https://github.com/RI-imaging/cellsino',
    version=version,
    packages=find_packages(),
    package_dir={name: name},
    include_package_data=True,
    license="MIT",
    description=description,
    long_description=open('README.rst').read() if exists('README.rst') else '',
    install_requires=["h5py>=2.7.0",
                      "qpimage",
                      "qpsphere>=0.5.0",
                      "numpy>=1.12.0",
                      ],
    setup_requires=['pytest-runner'],
    tests_require=["pytest"],
    python_requires='>=3.6, <4',
    keywords=["phase microscopy",
              "fluorescence imaging",
              "optical tomography",
              "diffraction tomography",
              ],
    classifiers= [
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Science/Research'
                 ],
    platforms=['ALL'],
    )



import numpy as np

from cellsino.elements.base_element import BaseElement


def test_translation():
    points = np.array([[0, 0, 0]])
    el = BaseElement(object_index=1.36,
                     medium_index=1.335,
                     fl_brightness=0,
                     points=points)

    el2 = el.transform(x=1, y=2, z=-3)
    assert np.all(el2.points[0] == [1, 2, -3])


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import abc

import numpy as np
import qpimage

from ..elements import Sphere


class BasePropagator(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, phantom, grid_size, pixel_size, wavelength):
        """Base propagator

        Notes
        -----
        - The origin of the coordinate system is the center of the grid.
          For even grid sizes, the origin is between two pixels. For odd
          grid sizes the origin coincides with the center pixel.
        """
        self.phantom = phantom
        self.wavelength = wavelength
        self.pixel_size = pixel_size
        self.grid_size = grid_size
        gx, gy = grid_size
        #: center of the volume used
        self.center = np.array([gx, gy, 0]) / 2 - .5

    def propagate(self):
        field = np.ones(self.grid_size, dtype=np.complex256)
        for element in self.phantom:
            if isinstance(element, Sphere):
                field *= self.propagate_sphere(element).field
        qpifull = qpimage.QPImage(data=field,
                                  which_data="field",
                                  meta_data={"wavelength": self.wavelength,
                                             "pixel size": self.pixel_size,
                                             }
                                  )
        return qpifull

    @abc.abstractmethod
    def propagate_sphere(self, sphere):
        pass



"""Reconstruction of a simple cell phantom

This example uses :ref:`ODTbrain<odtbrain:index>` and
:ref:`radontea<radontea:index>` for the reconstruction of the refractive
index and the fluorescence of the simple cell phantom.
The reconstruction is compared to the ground truth of the cell phantom.
"""
import cellsino
import matplotlib.pylab as plt
import numpy as np
import odtbrain as odt
import radontea as rt


# number of sinogram angles
num_ang = 160
# sinogram acquisition angles
angles = np.linspace(0, 2*np.pi, num_ang, endpoint=False)
# detector grid size
grid_size = (250, 250)
# vacuum wavelength [m]
wavelength = 550e-9
# pixel size [m]
pixel_size = 0.08e-6
# refractive index of the surrounding medium
medium_index = 1.335

# initialize cell phantom
phantom = cellsino.phantoms.SimpleCell()

# initialize sinogram with geometric parameters
sino = cellsino.Sinogram(phantom=phantom,
                         wavelength=wavelength,
                         pixel_size=pixel_size,
                         grid_size=grid_size)

# compute sinogram (field according to Rytov approximation and fluorescence)
sino_field, sino_fluor = sino.compute(angles=angles, propagator="rytov")

# reconstruction of refractive index
sino_rytov = odt.sinogram_as_rytov(sino_field)
potential = odt.backpropagate_3d(uSin=sino_rytov,
                                 angles=angles,
                                 res=wavelength/pixel_size,
                                 nm=medium_index)
ri = odt.odt_to_ri(f=potential,
                   res=wavelength/pixel_size,
                   nm=medium_index)

# reconstruction of fluorescence
fl = rt.backproject_3d(sinogram=sino_fluor,
                       angles=angles)

# reference for comparison
rimod, flmod = phantom.draw(grid_size=ri.shape,
                            pixel_size=pixel_size)

# plotting
idx = 150

plt.figure(figsize=(7, 5.5))

plotkwri = {"vmax": ri.real.max(),
            "vmin": ri.real.min(),
            "interpolation": "none",
            "cmap": "viridis",
            }

plotkwfl = {"vmax": fl.max(),
            "vmin": 0,
            "interpolation": "none",
            "cmap": "Greens_r",
            }

ax1 = plt.subplot(221, title="refractive index (ground_truth)")
mapper = ax1.imshow(rimod[idx].real, **plotkwri)
plt.colorbar(mappable=mapper, ax=ax1)

ax2 = plt.subplot(222, title="refractive index (reconstruction)")
mapper = ax2.imshow(ri[idx].real, **plotkwri)
plt.colorbar(mappable=mapper, ax=ax2)

ax3 = plt.subplot(223, title="fluorescence (ground truth)")
mapper = ax3.imshow(flmod[idx], **plotkwfl)
plt.colorbar(mappable=mapper, ax=ax3)

ax4 = plt.subplot(224, title="fluorescence (reconstruction)")
mapper = ax4.imshow(fl[idx], **plotkwfl)
plt.colorbar(mappable=mapper, ax=ax4)

for ax in [ax1, ax2, ax3, ax4]:
    ax.grid(color="w", alpha=.5)
    ax.set_xticks(np.arange(0, grid_size[0], 50))
    ax.set_yticks(np.arange(0, grid_size[0], 50))
    ax.set_xticklabels([])
    ax.set_yticklabels([])

plt.tight_layout()

plt.show()
